@@ -3,7 +3,9 @@ from ropy.robot.Link import Link
 from ropy.robot.Revolute import Revolute
 from ropy.robot.Prismatic import Prismatic
 from ropy.robot.ETS import ETS
+from ropy.robot.ELink import ELink
 from ropy.robot.ET import ET
+from ropy.robot.Shape import Shape
 
 __all__ = [
     'SerialLink',
@@ -11,5 +13,7 @@ __all__ = [
     'Revolute',
     'Prismatic',
     'ETS',
-    'ET'
+    'ELink',
+    'ET',
+    'Shape'
 ]
