@@ -1,7 +1,5 @@
 #!/usr/bin/env python3
-# -*- coding: utf-8 -*-
 """
-Created on Fri May 1 14:04:04 2020
 @author: Jesse Haviland
 """
 
@@ -10,9 +8,10 @@ import numpy as np
 import ropy as rp
 import spatialmath as sm
 import unittest
+import pytest
 
 
-class TestLink(unittest.TestCase):
+class TestSerialLink(unittest.TestCase):
 
     def test_seriallink(self):
         l0 = rp.Link()
@@ -974,11 +973,12 @@ class TestLink(unittest.TestCase):
 
         z = np.zeros(6)
         o = np.ones(6)
+        fext = [1, 2, 3, 1, 2, 3]
 
         tr0 = [-0.0000, 31.6399, 6.0351, 0.0000, 0.0283, 0]
         tr1 = [29.1421, 56.5044, 16.3528, 1.2645, 1.1239, 0.5196]
         tr2 = [32.4952, 60.8670, 17.7436, 1.4545, 1.2991, 0.7138]
-        # tr3 = [32.2416, 59.2850, 16.7723, 2.8687, 0.2991, 1.7138]
+        tr3 = [29.7849, 53.9511, 15.0208,  4.0929, -0.8761, 3.5196]
 
         t0 = puma.rne(z, z, puma.qn)
         t1 = puma.rne(z, o, puma.qn)
@@ -986,15 +986,13 @@ class TestLink(unittest.TestCase):
         puma.gravity = [0, 0, 9.81]
         t2 = puma.rne(o, o, puma.qn)
         t3 = puma.rne(z, z, grav=[0, 0, 9.81])
-        puma.rne(z, z, q=puma.qn, fext=o)
+        t4 = puma.rne(z, o, q=puma.qn, fext=fext)
 
         nt.assert_array_almost_equal(t0, tr0, decimal=4)
         nt.assert_array_almost_equal(t1, tr1, decimal=4)
         nt.assert_array_almost_equal(t2, tr2, decimal=4)
         nt.assert_array_almost_equal(t3, tr0, decimal=4)
-
-        # TODO Debug fext
-        # nt.assert_array_almost_equal(t4, tr3, decimal=4)
+        nt.assert_array_almost_equal(t4, tr3, decimal=4)
 
     def test_rne_traj(self):
         puma = rp.Puma560()
@@ -1174,3 +1172,331 @@ class TestLink(unittest.TestCase):
             "tool:  t = (0, 0, 0),  RPY/xyz = (0, 0, 0) deg")
 
         self.assertEqual(str(puma), res)
+
+    def test_paycap(self):
+        puma = rp.Puma560()
+        puma.q = puma.qn
+        q = puma.qn
+
+        w = [1, 2, 1, 2, 1, 2]
+        tauR = np.ones((6, 2))
+        tauR[:, 1] = -1
+
+        res0 = [
+            1.15865438e+00, -3.04790052e+02, -5.00870095e+01,  6.00479950e+15,
+            3.76356072e+00, 1.93649167e+00]
+
+        wmax0, joint = puma.paycap(w, tauR, q=q, frame=0)
+        wmax1, _ = puma.paycap(np.c_[w, w], tauR, q=np.c_[q, q], frame=0)
+        wmax2, _ = puma.paycap(w, tauR, frame=0)
+
+        nt.assert_allclose(wmax0, res0)
+        self.assertEqual(joint, 1)
+        nt.assert_allclose(wmax1[:, 0], res0)
+        nt.assert_allclose(wmax1[:, 1], res0)
+        nt.assert_allclose(wmax2, res0)
+
+    def test_jacob_dot(self):
+        puma = rp.Puma560()
+        puma.q = puma.qr
+        puma.qd = puma.qr
+        q = puma.qr
+
+        j0 = puma.jacob_dot(q, q)
+        j1 = puma.jacob_dot()
+
+        res = [-0.0000, -1.0654, -0.3702,  2.4674, 0, 0]
+
+        nt.assert_array_almost_equal(j0, res, decimal=4)
+        nt.assert_array_almost_equal(j1, res, decimal=4)
+
+    def test_yoshi(self):
+        puma = rp.Puma560()
+        puma.q = puma.qn
+        q = puma.qn
+
+        m0 = puma.maniplty()
+        m1 = puma.maniplty(q)
+        m2 = puma.maniplty(np.c_[q, q])
+        m3 = puma.maniplty(q, axes=[1, 1, 1, 0, 0, 0])
+        m4 = puma.maniplty(q, axes=[0, 0, 0, 1, 1, 1])
+
+        a0 = 0.0786
+        a2 = 0.111181
+        a3 = 2.44949
+
+        nt.assert_almost_equal(m0, a0, decimal=4)
+        nt.assert_almost_equal(m1, a0, decimal=4)
+        nt.assert_almost_equal(m2[0], a0, decimal=4)
+        nt.assert_almost_equal(m2[1], a0, decimal=4)
+        nt.assert_almost_equal(m3, a2, decimal=4)
+        nt.assert_almost_equal(m4, a3, decimal=4)
+
+    def test_asada(self):
+        puma = rp.Puma560()
+        puma.q = puma.qn
+        q = puma.qn
+
+        m0, mx0 = puma.maniplty(method='asada')
+        m1, mx1 = puma.maniplty(q, method='asada')
+        m2, mx2 = puma.maniplty(np.c_[q, q], method='asada')
+        m3, mx3 = puma.maniplty(q, axes=[1, 1, 1, 0, 0, 0], method='asada')
+        m4, mx4 = puma.maniplty(q, axes=[0, 0, 0, 1, 1, 1], method='asada')
+        m5, mx5 = puma.maniplty(puma.qz, method='asada')
+
+        a0 = 0.0044
+        a2 = 0.2094
+        a3 = 0.1716
+        a4 = 0.0
+
+        ax0 = np.array([
+            [17.2954, -2.7542, -9.6233, -0.0000,  0.2795, -0.0000],
+            [-2.7542, 12.1909,  1.2459, -0.3254, -0.0703, -0.9652],
+            [-9.6233,  1.2459, 13.3348, -0.0000,  0.2767,  0.0000],
+            [-0.0000, -0.3254, -0.0000,  0.1941,  0.0000,  0.1941],
+            [0.2795, -0.0703,  0.2767,  0.0000,  0.1713,  0.0000],
+            [-0.0000, -0.9652,  0.0000,  0.1941,  0.0000,  0.5791]
+        ])
+
+        ax1 = np.array([
+            [17.2954, -2.7542, -9.6233],
+            [-2.7542, 12.1909,  1.2459],
+            [-9.6233,  1.2459, 13.3348]
+        ])
+
+        ax2 = np.array([
+            [0.1941, 0.0000, 0.1941],
+            [0.0000, 0.1713, 0.0000],
+            [0.1941, 0.0000, 0.5791]
+        ])
+
+        ax3 = np.zeros((6, 6))
+
+        nt.assert_almost_equal(m0, a0, decimal=4)
+        nt.assert_array_almost_equal(mx0, ax0, decimal=4)
+        nt.assert_almost_equal(m1, a0, decimal=4)
+        nt.assert_array_almost_equal(mx1, ax0, decimal=4)
+        nt.assert_almost_equal(m2[0], a0, decimal=4)
+        nt.assert_array_almost_equal(mx2[:, :, 0], ax0, decimal=4)
+        nt.assert_almost_equal(m2[1], a0, decimal=4)
+        nt.assert_array_almost_equal(mx2[:, :, 1], ax0, decimal=4)
+
+        nt.assert_almost_equal(m3, a2, decimal=4)
+        nt.assert_array_almost_equal(mx3, ax1, decimal=4)
+
+        nt.assert_almost_equal(m4, a3, decimal=4)
+        nt.assert_array_almost_equal(mx4, ax2, decimal=4)
+
+        nt.assert_almost_equal(m5, a4, decimal=4)
+        nt.assert_array_almost_equal(mx5, ax3, decimal=4)
+
+    def test_maniplty_fail(self):
+        puma = rp.Puma560()
+        puma.q = puma.qn
+
+        with self.assertRaises(ValueError):
+            puma.maniplty(method='notamethod')
+
+    def test_perterb(self):
+        puma = rp.Puma560()
+        p2 = puma.perterb()
+        p3 = puma.perterb(0.8)
+
+        resI0 = np.zeros(puma.n)
+        resm0 = np.zeros(puma.n)
+        resI1 = np.zeros(puma.n)
+        resm1 = np.zeros(puma.n)
+
+        for i in range(puma.n):
+            resI0[i] = np.divide(
+                np.sum(np.abs(puma.links[i].I - p2.links[i].I)),
+                np.sum(np.abs(puma.links[i].I)))
+
+            if puma.links[i].m - p2.links[i].m != 0.0:
+                resm0[i] = np.abs(np.divide(
+                    (puma.links[i].m - p2.links[i].m),
+                    puma.links[i].m))
+            else:
+                resm0[i] = 0
+
+            resI1[i] = np.divide(
+                np.sum(np.abs(puma.links[i].I - p3.links[i].I)),
+                np.sum(np.abs(puma.links[i].I)))
+
+            if puma.links[i].m - p3.links[i].m != 0.0:
+                resm1[i] = np.abs(np.divide(
+                    (puma.links[i].m - p3.links[i].m),
+                    puma.links[i].m))
+            else:
+                resm1[i] = 0
+
+            self.assertTrue(resI0[i] < 0.1)
+            self.assertTrue(resm0[i] < 0.1 or np.isnan(resm0[i]))
+            self.assertTrue(resI1[i] < 0.8)
+            self.assertTrue(resm1[i] < 0.8 or np.isnan(resm1[i]))
+
+    def test_qmincon(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+
+        q = panda.qr
+        qt = np.c_[q, q]
+
+        q0, s0, _ = panda.qmincon()
+        q1, s1, _ = panda.qmincon(q)
+        q2, _, _ = panda.qmincon(qt)
+
+        qres = [-0.0969, -0.3000, 0.0870, -2.2000, 0.0297, 2.0000, 0.7620]
+
+        nt.assert_array_almost_equal(q0, qres, decimal=4)
+        nt.assert_array_almost_equal(q1, qres, decimal=4)
+        nt.assert_array_almost_equal(q2[:, 0], qres, decimal=4)
+        nt.assert_array_almost_equal(q2[:, 1], qres, decimal=4)
+
+    def test_teach(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.teach(block=False)
+        e.close()
+
+    def test_teach_withq(self):
+        panda = rp.PandaMDH()
+        e = panda.teach(q=panda.qr, block=False)
+        e.close()
+
+    def test_plot(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot(block=False)
+        e.close()
+
+    def test_teach_basic(self):
+        l0 = rp.Link(d=2)
+        r0 = rp.SerialLink([l0, l0])
+        e = r0.teach(False)
+        e.step()
+        e.close()
+
+    def test_plot_traj(self):
+        panda = rp.PandaMDH()
+        q = np.random.rand(7, 3)
+        e = panda.plot(block=False, q=q, dt=0)
+        e.close()
+
+    def test_control_type(self):
+        panda = rp.PandaMDH()
+
+        panda.control_type = 'p'
+
+        with self.assertRaises(ValueError):
+            panda.control_type = 'z'
+
+    def test_plot_vellipse(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+
+        e = panda.plot_vellipse(block=False, limits=[1, 2, 1, 2, 1, 2])
+        e.close()
+
+        e = panda.plot_vellipse(
+            block=False, q=panda.qr, centre='ee', opt='rot')
+        e.step(0)
+        e.close()
+
+        with self.assertRaises(TypeError):
+            panda.plot_vellipse(vellipse=10)
+
+        with self.assertRaises(ValueError):
+            panda.plot_vellipse(centre='ff')
+
+    def test_plot_fellipse(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+
+        e = panda.plot_fellipse(block=False, limits=[1, 2, 1, 2, 1, 2])
+        e.close()
+
+        e = panda.plot_fellipse(
+            block=False, q=panda.qr, centre='ee', opt='rot')
+        e.step(0)
+        e.close()
+
+        with self.assertRaises(TypeError):
+            panda.plot_fellipse(fellipse=10)
+
+        with self.assertRaises(ValueError):
+            panda.plot_fellipse(centre='ff')
+
+    def test_plot_with_vellipse(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot(block=False, vellipse=True)
+        e.close()
+
+    def test_plot_with_fellipse(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot(block=False, fellipse=True)
+        e.close()
+
+    def test_plot2(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot2(block=False, name=True)
+        e.close()
+
+    def test_plot2_traj(self):
+        panda = rp.PandaMDH()
+        q = np.random.rand(7, 3)
+        e = panda.plot2(block=False, q=q, dt=0)
+        e.close()
+
+    def test_teach2_basic(self):
+        l0 = rp.Link(d=2)
+        r0 = rp.SerialLink([l0, l0])
+        e = r0.teach2(False)
+        e.step()
+        e.close()
+
+    def test_teach2(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.teach(block=False)
+        e.close()
+
+        e2 = panda.teach2(block=False, q=panda.qr)
+        e2.close()
+
+    def test_plot_with_vellipse2(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot2(block=False, vellipse=True, limits=[1, 2, 1, 2])
+        e.step()
+        e.close()
+
+    def test_plot_with_fellipse2(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = panda.plot2(block=False, fellipse=True)
+        e.close()
+
+    def test_plot_with_vellipse2_fail(self):
+        panda = rp.PandaMDH()
+        panda.q = panda.qr
+        e = rp.backend.PyPlot2()
+        e.launch()
+        e.add(panda.fellipse(
+                q=panda.qr, centre=[0, 1]))
+        # e.close()
+
+        with self.assertRaises(ValueError):
+            e.add(panda.fellipse(
+                q=panda.qr, centre='ee', opt='rot'))
+
+        e.close()
+
+
+if __name__ == '__main__':
+
+    # unittest.main()
+    pytest.main(['tests/test_SerialLink.py'])
